import java.util.ArrayList;
import java.util.List;


import aima.search.framework.Successor;
import aima.search.framework.SuccessorFunction;


public class MCFunzioneSuccessore implements SuccessorFunction{

	@Override
	public List getSuccessors(Object stato) {
		
		List successori = new ArrayList();
		
		if (stato instanceof MCStato) {
			MCStato corrente = (MCStato)stato;
			
			int numCannibali;
			int numMissionari;
			
			//recupera il numero di M e C sulla sponda dove si trova la barca
			if (corrente.isBarca()) {
				//barca sulla sponda di partenza
				numCannibali = corrente.getCannibali();
				numMissionari = corrente.getMissionari();
			} else {
				//barca sulla sponda opposta
				numCannibali = 3 - corrente.getCannibali();
				numMissionari = 3 - corrente.getMissionari();
			}
			
			//Test mosse possibili tra le 2 sponde (partenza-->arrivo e viceversa)
			
			//se c' almeno 1C e 1M sulla sponda su cui si trova attualmente la barca (partenza o arrivo), posso eseguire
			//la mossa "muovi 1M e 1C", generando un 'nuovo' stato
			//si controlla se il nuovo stato  ammissibile prima di aggiungerlo alla lista degli stati successori
			//si aggiunge un'istanza della classe Successor: Successor(java.lang.String action, java.lang.Object state) che
			//tiene traccia del nuovo stato e dell'azione eseguita per raggiungerlo 
			if ((numCannibali>0) && (numMissionari>0)) {
				MCStato nuovo = muoviMC(corrente);
				if (consentito(nuovo)) successori.add(new Successor("MC", nuovo));
			}
			
			if (numCannibali>1) {
				MCStato nuovo = muoviCC(corrente);
				if (consentito(nuovo)) successori.add(new Successor("CC", nuovo));
			}
			
			if (numMissionari>1) {
				MCStato nuovo = muoviMM(corrente);
				if (consentito(nuovo)) successori.add(new Successor("MM", nuovo));
			}
			
			if (numCannibali>0) {
				MCStato nuovo = muoviC(corrente);
				if (consentito(nuovo)) successori.add(new Successor("C", nuovo));
			}
			
			if (numMissionari>0) {
				MCStato nuovo = muoviM(corrente);
				if (consentito(nuovo)) successori.add(new Successor("M", nuovo));
			}
		}
		return successori;
	}

	//i missionari devono prevalere sui cannibali
	private boolean consentito(MCStato stato) {
		if (stato.getMissionari() == 0 ||
			stato.getMissionari() == 3 ||
			(stato.getMissionari() >= stato.getCannibali() && 
					3-stato.getMissionari() >= 3-stato.getCannibali() ))
			// quest'ultima condizione serve per verificare che anche sull'altra sponda
			// ci sia uno stato ammissibile 			
			return true;
		else 
			return false;
	}
	
	//in ogni metodo "muovi" si considerano i 2 versi delle traversate (partenza-->arrivo e viceversa), controllando
	//dove si trova la barca nello stato corrente
	private MCStato muoviMM(MCStato stato) {
		if (stato.isBarca()) {
			//direzione sponda partenza --> sponda opposta (porto via 2 M)
			return new MCStato(stato.getMissionari()-2, stato.getCannibali(), !stato.isBarca());
		} else {
			//direzione sponda opposta --> sponda partenza (riporto indietro 2 M)
			return new MCStato(stato.getMissionari()+2, stato.getCannibali(), !stato.isBarca());
		}
	}
	
	private MCStato muoviCC(MCStato stato) {
		if (stato.isBarca()) {
			return new MCStato(stato.getMissionari(), stato.getCannibali()-2, !stato.isBarca());
		} else {
			return new MCStato(stato.getMissionari(), stato.getCannibali()+2, !stato.isBarca());
		}
	}
	
	private MCStato muoviMC(MCStato stato) {
		if (stato.isBarca()) {
			return new MCStato(stato.getMissionari()-1, stato.getCannibali()-1, !stato.isBarca());
		} else {
			return new MCStato(stato.getMissionari()+1, stato.getCannibali()+1, !stato.isBarca());
		}
	}
	
	private MCStato muoviC(MCStato stato) {
		if (stato.isBarca()) {
			return new MCStato(stato.getMissionari(), stato.getCannibali()-1, !stato.isBarca());
		} else {
			return new MCStato(stato.getMissionari(), stato.getCannibali()+1, !stato.isBarca());
		}
	}
	
	private MCStato muoviM(MCStato stato) {
		if (stato.isBarca()) {
			return new MCStato(stato.getMissionari()-1, stato.getCannibali(), !stato.isBarca());
		} else {
			return new MCStato(stato.getMissionari()+1, stato.getCannibali(), !stato.isBarca());
		}
	}
}
